import pickle
import pickletools
from typing import IO, Any

from .pickle import BuiltinPicklerIO
from .registry import register_pickler

__all__ = ["OptimizedPicklerIO"]


class OptimizedPicklerIO(BuiltinPicklerIO):
    """A PicklerIO class that combines :func:`pickletools.optimize` and standard :func:`pickle.dump`."""

    def dump(self, obj: Any, stream: IO[bytes], **kwargs):
        data = pickletools.optimize(pickle.dumps(obj, **kwargs))
        stream.write(data)


register_pickler("optimized_pickle", OptimizedPicklerIO)



from . import compressers, io, picklers, utils
from .compress_pickle import dump, dumps, load, loads
from .compressers.registry import (
    get_default_compression_mapping,
    get_known_compressions,
    get_registered_extensions,
)
from .picklers.registry import get_known_picklers

__version__ = "2.1.0"



from . import base, bz2, gzip, lz4, lzma, no_compression, registry, zipfile
from .base import BaseCompresser
from .bz2 import Bz2Compresser
from .gzip import GzipCompresser
from .lz4 import Lz4Compresser
from .lzma import LzmaCompresser
from .no_compression import NoCompresser
from .registry import *
from .zipfile import ZipfileCompresser



from abc import abstractmethod
from typing import IO, Any

__all__ = ["BasePicklerIO"]


class BasePicklerIO:
    """PicklerIO abstract base class.

    This class is in charge of writing and reading serialized python objects from a file-like
    binary stream.
    """

    @abstractmethod
    def dump(self, obj: Any, stream: IO[bytes], **kwargs):  # pragma: no cover
        """Write a serialized binary representation of an object to a stream.

        Parameters
        ----------
        obj : Any
            The python object that must be serialized and written.
        stream : IO[bytes]
            The binary stream (file-like object) where the serialized object must be written to.
        kwargs
            Any extra keyword arguments that are needed in the concrete implementation of ``dump``.
        """
        pass

    @abstractmethod
    def load(self, stream: IO[bytes], **kwargs):  # pragma: no cover
        """Load a serialized binary representation of an object from a stream.

        Parameters
        ----------
        stream : IO[bytes]
            The binary stream (file-like object) from where the serialized object must be loaded.
        kwargs
            Any extra keyword arguments that are needed in the concrete implementation of ``load``.

        Returns
        -------
        obj : Any
            The python object that was loaded.
        """
        pass



# -*- coding: utf-8 -*-
"""
A thin wrapper of standard ``pickle`` with standard compression libraries
"""
import io
import os
from typing import IO, Any, Dict, Optional, Union

from .io import compress_and_pickle, uncompress_and_unpickle
from .picklers import get_pickler
from .utils import instantiate_compresser

__all__ = ["dump", "load", "dumps", "loads"]


PathLike = os.PathLike
PathType = Union[str, bytes, PathLike]
FileType = IO[bytes]


def dump(
    obj: Any,
    path: Union[PathType, FileType],
    compression: Optional[str] = "infer",
    pickler_method: str = "pickle",
    pickler_kwargs: Optional[Dict[str, Any]] = None,
    mode: Optional[str] = None,
    *,
    set_default_extension: bool = True,
    **kwargs,
):
    """Serialize an object and write it to a file-like object.

    It is possible to specify a desired serialization method and a compression protocol to use.
    For example, if ``gzip`` compression is specified, the object is serialized using the standard
    :func:`pickle.dump` and compressed using gzip.

    Behind the scenes, this function only instantiates
    :class:`~compress_pickle.compressers.base.BaseCompresser` and
    :class:`~compress_pickle.picklers.base.BasePicklerIO` instances depending on the supplied
    names, and then calls :func:`~compress_pickle.io.base.compress_and_pickle` with them.

    Parameters
    ----------
    obj : Any
        The object that will be serialized
    path : Union[PathType, FileType]
        A path-like object (``str``, ``bytes``, ``os.PathType``) or a file-like object
        (``io.BaseIO`` instances). The path to which to dump the ``obj``.
    compression : Optional[str]
        The compression protocol to use. By default, the compression is inferred from the path's
        extension. This compression name passed to
        :func:`~compress_pickle.compressers.registry.get_compresser` to create a
        :class:`~compress_pickle.compressers.base.BaseCompresser` instance. This instance will be
        used to create a file-like object onto which to write the serialized binary representation
        of ``obj``.
        To see available compression protocols refer to
        :func:`~compress_pickle.compressers.registry.get_known_compressions`.
    pickler_method : str
        The name of the serialization method to use. This method name is passed to
        :func:`~compress_pickle.picklers.registry.get_pickler` to create a
        :class:`~compress_pickle.picklers.base.BasePicklerIO` instance. The default method is
        ``"pickle"``, which means that a :class:`~compress_pickle.picklers.pickle.BuiltinPicklerIO`
        will be created by default, and its
        :meth:`~compress_pickle.picklers.pickle.BuiltinPicklerIO.dump` will be used for
        serializing the ``obj``.
        Refer to :func:`~compress_pickle.picklers.registry.get_known_picklers` to see the available
        serialization methods.
    pickler_kwargs : Optional[Dict[str, Any]]
        An optional dictionary of keyword arguments to pass to
        :meth:`~compress_pickle.picklers.base.BasePicklerIO.dump` when ``obj`` is serialized. For
        example, this could be ``{"protocol": -1, "fix_imports": True}`` when the default
        ``"pickle"`` serialization method is used.
    mode : Optional[str]
        Mode with which to open the file buffer. The default changes according to the compression
        protocol. Refer to :func:`~compress_pickle.compressers.registry.get_compression_write_mode`
        to see the defaults.
    set_default_extension : bool
        If ``True``, the default extension given the provided compression protocol is set to the
        supplied ``path``. Refer to
        :func:`~compress_pickle.compressers.registry.get_default_compression_mapping` for the
        default extension registered to each compression method.
    kwargs :
        Any extra keyword arguments are passed to
        :func:`compress_pickle.utils.instantiate_compresser`, which in turn creates the
        compresser instance that will be used.

    Notes
    -----
    If the supplied ``path`` is a file-like object, ``dump`` does not close it before exiting.
    The users must handle the closing on their own. If the supplied ``path`` is a path-like object,
    ``dump`` opens and then closes the file after the writting process finishes.
    """
    _mode = "write" if mode is None else mode
    pickler = get_pickler(pickler_method)()
    compresser = instantiate_compresser(
        compression=compression,
        path=path,
        mode=_mode,
        set_default_extension=set_default_extension,
        **kwargs,
    )
    if pickler_kwargs is None:
        pickler_kwargs = {}
    try:
        compress_and_pickle(
            compresser,
            pickler=pickler,
            obj=obj,
            **pickler_kwargs,
        )
    finally:
        compresser.close()


def dumps(
    obj: Any,
    compression: Optional[str],
    pickler_method: str = "pickle",
    pickler_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> bytes:
    """Serialize an object and return its binary representation.

    It is possible to specify a desired serialization method and a compression protocol to use.
    For example, if ``gzip`` compression is specified, the serialized object is compressed using
    gzip.

    Behind the scenes, this function only instantiates
    :class:`~compress_pickle.compressers.base.BaseCompresser` around a ``io.BinaryIO`` and
    :class:`~compress_pickle.picklers.base.BasePicklerIO` instances depending on the supplied
    names, and then calls :func:`~compress_pickle.io.base.compress_and_pickle` with them. The
    contents of the ``io.BinaryIO`` are then returned.

    Parameters
    ----------
    obj : Any
        The object that will be serialized.
    compression : Optional[str]
        The compression protocol to use. By default, the compression is inferred from the path's
        extension. This compression name passed to
        :func:`~compress_pickle.compressers.registry.get_compresser` to create a
        :class:`~compress_pickle.compressers.base.BaseCompresser` instance. This instance will be
        used to create a file-like object onto which to write the serialized binary representation
        of ``obj``.
        To see available compression protocols refer to
        :func:`~compress_pickle.compressers.registry.get_known_compressions`.
    pickler_method : str
        The name of the serialization method to use. This method name is passed to
        :func:`~compress_pickle.picklers.registry.get_pickler` to create a
        :class:`~compress_pickle.picklers.base.BasePicklerIO` instance. The default method is
        ``"pickle"``, which means that a :class:`~compress_pickle.picklers.pickle.BuiltinPicklerIO`
        will be created by default, and its
        :meth:`~compress_pickle.picklers.pickle.BuiltinPicklerIO.dump` will be used for
        serializing the ``obj``.
        Refer to :func:`~compress_pickle.picklers.registry.get_known_picklers` to see the available
        serialization methods.
    pickler_kwargs : Optional[Dict[str, Any]]
        An optional dictionary of keyword arguments to pass to
        :meth:`~compress_pickle.picklers.base.BasePicklerIO.dump` when ``obj`` is serialized. For
        example, this could be ``{"protocol": -1, "fix_imports": True}`` when the default
        ``"pickle"`` serialization method is used.

    Returns
    -------
    bytes
        The resulting bytes of the serialized ``obj`` after being compressed.

    Notes
    -----
    The ``compression`` argument is mandatory because it cannot be inferred.
    """
    with io.BytesIO() as stream:
        dump(
            obj,
            path=stream,
            compression=compression,
            pickler_method=pickler_method,
            pickler_kwargs=pickler_kwargs,
            **kwargs,
        )
        return stream.getvalue()


def load(
    path: Union[PathType, FileType],
    compression: Optional[str] = "infer",
    pickler_method: str = "pickle",
    pickler_kwargs: Optional[Dict[str, Any]] = None,
    mode: Optional[str] = None,
    *,
    set_default_extension: bool = True,
    **kwargs,
) -> Any:
    """Load an object from its serialized binary representation stored in a file-like object.

    It is possible to specify a desired serialization method and a compression protocol to use.
    For example, if ``gzip`` compression is specified, the contents of the supplied file-like
    object are uncompressed using gzip and then loaded using standard :func:`pickle.load`.

    Behind the scenes, this function only instantiates
    :class:`~compress_pickle.compressers.base.BaseCompresser` and
    :class:`~compress_pickle.picklers.base.BasePicklerIO` instances depending on the supplied
    names, and then calls :func:`~compress_pickle.io.base.uncompress_and_unpickle` with them.

    Parameters
    ----------
    path : Union[PathType, FileType]
        A path-like object (``str``, ``bytes``, ``os.PathType``) or a file-like object
        (``io.BaseIO`` instances). The path to which to dump the ``obj``.
    compression : Optional[str]
        The compression protocol to use. By default, the compression is inferred from the path's
        extension. This compression name passed to
        :func:`~compress_pickle.compressers.registry.get_compresser` to create a
        :class:`~compress_pickle.compressers.base.BaseCompresser` instance. This instance will be
        used to open a file-like object from which to read the serialized binary representation
        of ``obj``.
        To see available compression protocols refer to
        :func:`~compress_pickle.compressers.registry.get_known_compressions`.
    pickler_method : str
        The name of the serialization method to use. This method name is passed to
        :func:`~compress_pickle.picklers.registry.get_pickler` to create a
        :class:`~compress_pickle.picklers.base.BasePicklerIO` instance. The default method is
        ``"pickle"``, which means that a :class:`~compress_pickle.picklers.pickle.BuiltinPicklerIO`
        will be created by default, and its
        :meth:`~compress_pickle.picklers.pickle.BuiltinPicklerIO.load` will be used for
        unserializing the ``obj``.
        Refer to :func:`~compress_pickle.picklers.registry.get_known_picklers` to see the available
        serialization methods.
    pickler_kwargs : Optional[Dict[str, Any]]
        An optional dictionary of keyword arguments to pass to
        :meth:`~compress_pickle.picklers.base.BasePicklerIO.load` when the ``obj`` is loaded. For
        example, this could be ``{"fix_imports": True}`` when the default ``"pickle"``
        serialization method is used.
    mode : Optional[str]
        Mode with which to open the file buffer. The default changes according to the compression
        protocol. Refer to :func:`~compress_pickle.compressers.registry.get_compression_read_mode`
        to see the defaults.
    set_default_extension : bool
        If ``True``, the default extension given the provided compression protocol is set to the
        supplied ``path``. Refer to
        :func:`~compress_pickle.compressers.registry.get_default_compression_mapping` for the
        default extension registered to each compression method.
    kwargs :
        Any extra keyword arguments are passed to
        :func:`compress_pickle.utils.instantiate_compresser`, which in turn creates the
        compresser instance that will be used.

    Returns
    -------
    obj : Any
        The object that is loaded from its compressed binary representation.

    Notes
    -----
    If the supplied ``path`` is a file-like object, ``load`` does not close it before exiting.
    The users must handle the closing on their own. If the supplied ``path`` is a path-like object,
    ``load`` opens and then closes the file after the writting process finishes.
    """
    _mode = "read" if mode is None else mode
    pickler = get_pickler(pickler_method)()
    compresser = instantiate_compresser(
        compression=compression,
        path=path,
        mode=_mode,
        set_default_extension=set_default_extension,
        **kwargs,
    )
    if pickler_kwargs is None:
        pickler_kwargs = {}
    try:
        output = uncompress_and_unpickle(
            compresser,
            pickler=pickler,
            **pickler_kwargs,
        )
    finally:
        compresser.close()
    return output


def loads(
    data: bytes,
    compression: Optional[str],
    pickler_method: str = "pickle",
    pickler_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Any:
    """Load an object from its serialized binary representation.

    It is possible to specify a desired serialization method and a compression protocol to use.
    For example, if ``gzip`` compression is specified, the supplied binary data is uncompressed with
    gzip and then loaded using standard :func:`pickle.load`.

    Behind the scenes, this function only instantiates
    :class:`~compress_pickle.compressers.base.BaseCompresser` around a ``io.BinaryIO(data)`` stream
    and :class:`~compress_pickle.picklers.base.BasePicklerIO` instances depending on the supplied
    names, and then calls :func:`~compress_pickle.io.base.uncompress_and_unpickle` with them.

    Parameters
    ----------
    data : bytes
        The bytes that contain the compressed serialized binary representation of the object that
        must be loaded.
    compression : Optional[str]
        The compression protocol to use. By default, the compression is inferred from the path's
        extension. This compression name passed to
        :func:`~compress_pickle.compressers.registry.get_compresser` to create a
        :class:`~compress_pickle.compressers.base.BaseCompresser` instance. This instance will be
        used to open a file-like object from which to read the serialized binary representation
        of ``obj``.
        To see available compression protocols refer to
        :func:`~compress_pickle.compressers.registry.get_known_compressions`.
    pickler_method : str
        The name of the serialization method to use. This method name is passed to
        :func:`~compress_pickle.picklers.registry.get_pickler` to create a
        :class:`~compress_pickle.picklers.base.BasePicklerIO` instance. The default method is
        ``"pickle"``, which means that a :class:`~compress_pickle.picklers.pickle.BuiltinPicklerIO`
        will be created by default, and its
        :meth:`~compress_pickle.picklers.pickle.BuiltinPicklerIO.load` will be used for
        unserializing the ``obj``.
        Refer to :func:`~compress_pickle.picklers.registry.get_known_picklers` to see the available
        serialization methods.
    pickler_kwargs : Optional[Dict[str, Any]]
        An optional dictionary of keyword arguments to pass to
        :meth:`~compress_pickle.picklers.base.BasePicklerIO.load` when the ``obj`` is loaded. For
        example, this could be ``{"fix_imports": True}`` when the default ``"pickle"``
        serialization method is used.
    kwargs :
        Any extra keyword arguments are passed to
        :func:`compress_pickle.utils.instantiate_compresser`, which in turn creates the
        compresser instance that will be used.

    Returns
    -------
    obj : Any
        The object that is loaded from its compressed binary representation.

    Notes
    -----
    The ``compression`` argument is mandatory because it cannot be inferred.
    """
    with io.BytesIO(bytes(data)) as stream:
        return load(
            stream,
            compression=compression,
            pickler_method=pickler_method,
            pickler_kwargs=pickler_kwargs,
            **kwargs,
        )



import codecs
import re
from os import path

from setuptools import setup, find_packages


PROJECT_ROOT = path.dirname(path.abspath(__file__))
REQUIREMENTS_FILE = path.join(PROJECT_ROOT, "requirements.txt")
README_FILE = path.join(PROJECT_ROOT, "README.md")
VERSION_FILE = path.join(PROJECT_ROOT, "compress_pickle", "__init__.py")

NAME = "compress_pickle"
DESCRIPTION = "Standard pickle, wrapped with standard compression libraries"
AUTHOR = "Luciano Paz"
AUTHOR_EMAIL = "luciano.paz.neuro@gmail.com"
URL = "https://github.com/lucianopaz/compress_pickle"


lz4_requires = ["lz4"]
dill_requires = ["dill"]
cloudpickle_requires = ["cloudpickle"]
extras_require = {
    "lz4": lz4_requires,
    "dill": dill_requires,
    "cloudpickle": cloudpickle_requires,
    "full": lz4_requires + dill_requires + cloudpickle_requires,
}


def get_requirements():
    with codecs.open(REQUIREMENTS_FILE) as buff:
        return buff.read().splitlines()


def get_long_description():
    with codecs.open(README_FILE, "rt") as buff:
        return buff.read()


def get_version():
    lines = open(VERSION_FILE, "rt").readlines()
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in lines:
        mo = re.search(version_regex, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError("Unable to find version in %s." % (VERSION_FILE,))


if __name__ == "__main__":
    setup(
        name=NAME,
        version=get_version(),
        description=DESCRIPTION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        python_requires=">=3.6",
        packages=find_packages(),
        install_requires=get_requirements(),
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        include_package_data=True,
        extras_require=extras_require,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
    )



from functools import singledispatch
from typing import Any

from ..compressers.base import BaseCompresser
from ..picklers.base import BasePicklerIO

__all__ = ["compress_and_pickle", "uncompress_and_unpickle"]


@singledispatch
def compress_and_pickle(compresser: Any, pickler: BasePicklerIO, obj: Any, **kwargs):
    """Take an object, serialize it and write it to a compresser's stream.

    This is the main serialization function. It works as a singledispatch function, and its
    implementation is dependent on the type of the supplied ``compresser``. The default
    implementation only supports :class:`~compress_pickle.compressers.base.BaseCompresser`
    instances.

    The ``compresser`` is used to get the stream onto which to write the serialized ``obj``.
    The ``pickler`` instance is the object that is responsible for `dumping` the ``obj`` to the
    compresser's stream.

    Parameters
    ----------
    compresser : BaseCompresser
        The compresser instance that provides the file-like object onto which the ``obj`` must be
        written.
    pickler : BasePicklerIO
        The object that is responsible for serializing the ``obj`` and writting its binary
        representation into the file-like stream provided by the ``compresser``.
    obj : Any
        The object that you wish to serialize, compress and write.
    **kwargs
        Any extra keyword arguments are passed to the pickler's ``dump`` method.

    Raises
    ------
    NotImplementedError
        If the ``compresser`` is not an instance of
        :class:`~compress_pickle.compressers.base.BaseCompresser`
    """
    raise NotImplementedError(
        f"compress_and_pickle is not implemented for the supplied compresser type: "
        f"{type(compresser)}"
    )


@compress_and_pickle.register(BaseCompresser)
def default_compress_and_pickle(
    compresser: BaseCompresser, pickler: BasePicklerIO, obj: Any, **kwargs
):
    pickler.dump(obj=obj, stream=compresser.get_stream(), **kwargs)


@singledispatch
def uncompress_and_unpickle(compresser: Any, pickler: BasePicklerIO, **kwargs) -> Any:
    """Load and uncompress an object from a compresser's stream.

    This is the main loading function. It works as a singledispatch function, and its
    implementation is dependent on the type of the supplied ``compresser``. The default
    implementation only supports :class:`~compress_pickle.compressers.base.BaseCompresser`
    instances.

    The ``compresser`` is used to get the stream from which to load the serialized ``obj``.
    The ``pickler`` instance is the object that is responsible for `loading` the ``obj`` from the
    compresser's stream.

    Parameters
    ----------
    compresser : BaseCompresser
        The compresser instance that provides the file-like object from which to load the ``obj``.
    pickler : BasePicklerIO
        The object that is responsible for reading the contents of the file-like stream taken from
        the ``compresser`` and loading the serialized object
    **kwargs
        Any extra keyword arguments are passed to the pickler's ``load`` method.

    Returns
    -------
    Any
        The resulting uncompressed and unserialized object.

    Raises
    ------
    NotImplementedError
        If the ``compresser`` is not an instance of
        :class:`~compress_pickle.compressers.base.BaseCompresser`
    """
    raise NotImplementedError(
        f"uncompress_and_unpickle is not implemented for the supplied compresser type: "
        f"{type(compresser)}"
    )


@uncompress_and_unpickle.register(BaseCompresser)
def default_uncompress_and_unpickle(
    compresser: BaseCompresser, pickler: BasePicklerIO, **kwargs
) -> Any:
    return pickler.load(stream=compresser.get_stream(), **kwargs)



from typing import Dict, List, Optional, Sequence, Type

from .base import BaseCompresser

__all__ = [
    "get_compresser",
    "get_compresser_from_extension",
    "get_compression_from_extension",
    "register_compresser",
    "get_registered_extensions",
    "get_compression_write_mode",
    "get_compression_read_mode",
    "add_compression_alias",
    "get_known_compressions",
    "validate_compression",
    "get_default_compression_mapping",
    "list_registered_compressers",
]


class _compresser_registry:
    _compresser_registry: Dict[Optional[str], Type[BaseCompresser]] = {}

    _compresser_default_write_modes: Dict[Optional[str], str] = {}

    _compresser_default_read_modes: Dict[Optional[str], str] = {}

    _compression_extension_map: Dict[str, Optional[str]] = {}

    _compression_aliases: Dict[str, Optional[str]] = {}

    @classmethod
    def get_compresser(cls, compression: Optional[str]) -> Type[BaseCompresser]:
        """Get the compresser class registered with a given compression name.

        Parameters
        ----------
        compression : Optional[str]
            The compression name.

        Raises
        ------
        ValueError
            If the supplied ``compression`` has not been registered.

        Returns
        -------
        Type[BaseCompresser]
            The compresser class associated to the ``compression`` name.
        """
        try:
            return cls._compresser_registry[compression]
        except Exception:
            raise ValueError(
                f"Unknown compresser {compression}. "
                f"Available values are {list(cls._compresser_registry)}"
            )

    @classmethod
    def get_compresser_from_extension(cls, extension: str) -> Type[BaseCompresser]:
        """Get the compresser class registered with a given file extension.

        Parameters
        ----------
        extension : str
            The file extension, for example ".zip".
            Note that the dot characters will be striped from the left of any supplied extension
            before the lookup it. This means that ".zip" and "zip" will be considered equivalent
            extensions.

        Raises
        ------
        ValueError
            If the supplied ``extension`` has not been registered.

        Returns
        -------
        Type[BaseCompresser]
            The compresser class associated to the extension.
        """
        return cls._compresser_registry[cls.get_compression_from_extension(extension)]

    @classmethod
    def get_compression_from_extension(cls, extension: str) -> Optional[str]:
        """Get the compression name registered with a given file extension.

        Parameters
        ----------
        extension : str
            The file extension, for example ".zip".
            Note that the dot characters will be striped from the left of any supplied extension
            before the lookup it. This means that ".zip" and "zip" will be considered equivalent
            extensions.

        Raises
        ------
        ValueError
            If the supplied ``extension`` has not been registered.

        Returns
        -------
        Optional[str]
            The compression name associated to the extension.
        """
        try:
            return cls._compression_extension_map[extension.lstrip(".")]
        except Exception:
            raise ValueError(
                f"Unregistered extension {extension}. "
                f"Registered extensions are {list(cls._compression_extension_map)}"
            )

    @classmethod
    def register_compresser(
        cls,
        compression: Optional[str],
        compresser: Type[BaseCompresser],
        extensions: Sequence[str],
        default_write_mode: str = "wb",
        default_read_mode: str = "rb",
    ):
        """Register a compression method, along with its compresser class, extensions and modes.

        Parameters
        ----------
        compression : Optional[str]
            The compression name that will be registered.
        compresser : Type[BaseCompresser]
            The compresser class. This should be a :class:`~compress_pickle.compressers.base.BaseCompresser`
            subclass.
        extensions : Sequence[str]
            A sequence of file name extensions (e.g. [".zip"]) that will be registered to the
            supplied compression. These extensions will be used to infer the compression method to
            use from a file name. The first entry in the ``extensions`` sequence will be used as
            the compression's default extension name. For example, if ``extensions = ["bz2", "bz"]``
            both the extensions ``"bz2"`` and ``"bz"`` will be registered to the ``compression``,
            but ``"bz2"`` will be taken as the compression's default extension.
            Note that the dot characters will be striped from the left of any supplied extension
            before registering it. This means that ".zip" and "zip" will be considered equivalent
            extensions.
        default_write_mode : str
            The write mode with which to open the file object stream by default.
        default_read_mode : str
            The read mode with which to open the file object stream by default.

        Raises
        ------
        ValueError
            If the supplied ``compression`` is already contained in the registry or if any of the
            supplied extensions is already registered.
        TypeError
            If the supplied compresser is not a :class:`~compress_pickle.compressers.base.BaseCompresser`
            subclass.
        """
        if compression in cls._compresser_registry:
            raise ValueError(
                f"A compresser with name {compression} is already registered. "
                "Please choose a different name."
            )
        try:
            _subclass = issubclass(compresser, BaseCompresser)
            if not _subclass:
                raise TypeError()
        except Exception:
            raise TypeError(
                f"The supplied compresser {compresser} is not a derived from {BaseCompresser}"
            )
        extensions = [ext.lstrip(".") for ext in extensions]
        for ext in extensions:
            if ext in cls._compression_extension_map:
                raise ValueError(
                    f"Tried to register the extension {ext} to compresser {compression}, but it "
                    "is already registered in favour of compresser "
                    f"{cls._compression_extension_map[ext]}. Please use a different extension "
                    "instead."
                )
        cls._compresser_registry[compression] = compresser
        cls._compression_extension_map.update({ext: compression for ext in extensions})
        cls._compresser_default_write_modes[compression] = default_write_mode
        cls._compresser_default_read_modes[compression] = default_read_mode

    @classmethod
    def get_compression_write_mode(cls, compression: Optional[str]) -> str:
        """Get the compression's default mode for openning the file buffer for writing.

        Parameters
        ----------
        compression : Optional[str]
            The compression name.

        Returns
        -------
        compression_write_mode : str
            The default write mode of the given ``compression``.

        Raises
        ------
        ValueError
            If the default write mode of the supplied ``compression`` is not known.
        """
        try:
            return cls._compresser_default_write_modes[compression]
        except Exception:  # pragma: no cover
            raise ValueError(
                "Unknown compression {}. Available values are: {}".format(
                    compression, list(cls._compresser_default_write_modes)
                )
            )

    @classmethod
    def get_compression_read_mode(cls, compression: Optional[str]) -> str:
        """Get the compression's default mode for openning the file buffer for reading.

        Parameters
        ----------
        compression : Optional[str]
            The compression name.

        Returns
        -------
        compression_read_mode : str
            The default read mode of the given ``compression``.

        Raises
        ------
        ValueError
            If the default write mode of the supplied ``compression`` is not known.
        """
        try:
            return cls._compresser_default_read_modes[compression]
        except Exception:  # pragma: no cover
            raise ValueError(
                "Unknown compression {}. Available values are: {}".format(
                    compression, list(cls._compresser_default_read_modes)
                )
            )

    @classmethod
    def add_compression_alias(cls, alias: str, compression: Optional[str]):
        """Add an alias for an already registered compression.

        Parameters
        ----------
        alias : str
            The alias to register
        compression : Optional[str]
            The compression name that must already be registered.

        Raises
        ------
        ValueError
            If the supplied ``compression`` is not known or if the supplied ``alias``
            is already contained in the registry.
        """
        if alias in cls._compresser_registry:
            raise ValueError(
                f"The alias {alias} is already registered, please choose a different alias."
            )
        if compression not in cls._compresser_registry:
            raise ValueError(
                "Unknown compression {}. Available values are: {}".format(
                    compression, list(cls._compresser_registry)
                )
            )
        cls._compresser_registry[alias] = cls._compresser_registry[compression]
        cls._compresser_default_write_modes[
            alias
        ] = cls._compresser_default_write_modes[compression]
        cls._compresser_default_read_modes[alias] = cls._compresser_default_read_modes[
            compression
        ]
        cls._compression_aliases[alias] = compression


get_compresser = _compresser_registry.get_compresser

get_compresser_from_extension = _compresser_registry.get_compresser_from_extension

get_compression_from_extension = _compresser_registry.get_compression_from_extension

register_compresser = _compresser_registry.register_compresser

get_compression_read_mode = _compresser_registry.get_compression_read_mode

get_compression_write_mode = _compresser_registry.get_compression_write_mode

add_compression_alias = _compresser_registry.add_compression_alias


def get_registered_extensions() -> Dict[str, Optional[str]]:  # pragma: no cover
    """Get a copy of the mapping between file extensions and registered compressers.

    Returns
    -------
    Dict[str, Optional[str]]
        The mapping between file extensions and registered compressers.
    """
    return _compresser_registry._compression_extension_map.copy()


def get_known_compressions() -> List[Optional[str]]:
    """Get a list of known compression protocols

    Returns
    -------
    compressions : List[Optional[str]]
        List of known compression protocol names.
    """
    return list(_compresser_registry._compresser_registry)


def validate_compression(compression: Optional[str], infer_is_valid: bool = True):
    """Check if the supplied ``compression`` protocol is supported.

    If it is not supported, a ``ValueError`` is raised.

    Parameters
    ----------
    compression : Optional[str]
        A compression protocol. To see the known compression protocolos, use
        :func:`~.get_known_compressions`
    infer_is_valid : bool
        If ``True``, ``compression="infer"`` is considered a valid compression
        protocol. If ``False``, it is not accepted as a valid compression
        protocol.

    Raises
    ------
    ValueError
        If the supplied ``compression`` is not supported.
    """
    if compression in get_known_compressions():
        return True
    elif infer_is_valid and compression == "infer":
        return True
    raise ValueError(
        "Unknown compression {}. Available values are: {}".format(
            compression, get_known_compressions()
        )
    )


def get_default_compression_mapping() -> Dict[Optional[str], str]:
    """Get a mapping from known compression protocols to the default filename extensions.

    Returns
    -------
    compression_map : Dict[Optional[str], str]
        Dictionary that maps known compression protocol names to their default
        file extension.
    """
    output = {}
    for (
        extension,
        compresser_name,
    ) in _compresser_registry._compression_extension_map.items():
        if compresser_name not in output:
            output[compresser_name] = extension
    output.update(
        {
            alias: output[compression]
            for alias, compression in _compresser_registry._compression_aliases.items()
        }
    )
    return output


def list_registered_compressers() -> List[Type[BaseCompresser]]:  # pragma: no cover
    """Get the list of registered compresser classes.

    Returns
    -------
    List[Type[BaseCompresser]]
        The list of registered compresser classes.
    """
    return list(_compresser_registry._compresser_registry.values())
